from __future__ import annotations

from urllib.parse import urlencode

from allauth.account.adapter import DefaultAccountAdapter
from allauth.exceptions import ImmediateHttpResponse
from allauth.socialaccount.models import SocialLogin
from django.core.exceptions import PermissionDenied
from django.http import HttpRequest
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.urls import reverse

from allauth_2fa.utils import user_has_valid_totp_device


class OTPAdapter(DefaultAccountAdapter):
    def has_2fa_enabled(self, user) -> bool:
        """Returns True if the user has 2FA configured."""
        return user_has_valid_totp_device(user)

    def pre_login(
        self,
        request: HttpRequest,
        user,
        **kwargs,
    ) -> HttpResponse | None:
        response = super().pre_login(request, user, **kwargs)
        if response:
            return response

        # Require two-factor authentication if it has been configured.
        if self.has_2fa_enabled(user):
            self.stash_pending_login(request, user, kwargs)
            redirect_url = self.get_2fa_authenticate_url(request)
            raise ImmediateHttpResponse(response=HttpResponseRedirect(redirect_url))

        # Otherwise defer to the original allauth adapter.
        return super().login(request, user)

    def get_2fa_authenticate_url(self, request: HttpRequest) -> str:
        """
        Get the URL to redirect to for finishing 2FA authentication.
        """
        redirect_url = reverse("two-factor-authenticate")

        # Add "next" parameter to the URL if possible.
        # If the view function smells like a class-based view, we can interrogate it.
        if getattr(request.resolver_match.func, "view_class", None):
            view = request.resolver_match.func.view_class()
            view.request = request
            success_url = view.get_success_url()
            query_params = request.GET.copy()
            if success_url:
                query_params[view.redirect_field_name] = success_url
            if query_params:
                redirect_url += f"?{urlencode(query_params)}"

        return redirect_url

    def stash_pending_login(
        self,
        request: HttpRequest,
        user,
        login_kwargs: dict,
    ) -> None:
        """Here, we're going to stash the pending login so that it can be
        resumed at a later point. The `login_kwargs` contain meta information on
        the login which we need to store.

        The `login_kwargs` passed is a dictionary, and we're going to store that
        in the session. While doing so, we need to mutate it a bit to ensure
        serializability.  Mutating a dictionary passed to us is not something
        the caller expects, so we're going to make a shallow copy to prevent the
        caller from being impacted. Shallow is fine, as we're only setting new
        keys and not altering existing values.
        """
        # Cast to string for the case when this is not a JSON serializable
        # object, e.g. a UUID.
        request.session["allauth_2fa_user_id"] = str(user.id)
        login_kwargs = login_kwargs.copy()
        signal_kwargs = login_kwargs.get("signal_kwargs")
        if signal_kwargs:
            sociallogin = signal_kwargs.get("sociallogin")
            if sociallogin:
                signal_kwargs = signal_kwargs.copy()
                signal_kwargs["sociallogin"] = sociallogin.serialize()
                login_kwargs["signal_kwargs"] = signal_kwargs
        request.session["allauth_2fa_login"] = login_kwargs

    def unstash_pending_login_kwargs(self, request: HttpRequest) -> dict:
        login_kwargs = request.session.pop("allauth_2fa_login", None)
        if login_kwargs is None:
            raise PermissionDenied()
        signal_kwargs = login_kwargs.get("signal_kwargs")
        if signal_kwargs:
            sociallogin = signal_kwargs.get("sociallogin")
            if sociallogin:
                signal_kwargs["sociallogin"] = SocialLogin.deserialize(sociallogin)
        return login_kwargs



from __future__ import annotations

from django.contrib.auth.mixins import AccessMixin
from django.http import HttpRequest
from django.http import HttpResponse
from django.http import HttpResponseRedirect
from django.urls import reverse_lazy

from allauth_2fa.utils import user_has_valid_totp_device


class ValidTOTPDeviceRequiredMixin(AccessMixin):
    no_valid_totp_device_url = reverse_lazy("two-factor-setup")

    def dispatch(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        if not request.user.is_authenticated:
            return self.handle_no_permission()
        if not user_has_valid_totp_device(request.user):
            return self.handle_missing_totp_device()
        return super().dispatch(request, *args, **kwargs)

    def handle_missing_totp_device(self) -> HttpResponse:
        return HttpResponseRedirect(self.no_valid_totp_device_url)



from __future__ import annotations

from base64 import b64encode

from allauth.account.adapter import get_adapter
from allauth.utils import get_form_class
from django.contrib.auth import get_user_model
from django.contrib.auth.mixins import LoginRequiredMixin
from django.contrib.sites.shortcuts import get_current_site
from django.http import HttpResponseRedirect
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.encoding import force_str
from django.views.generic import FormView
from django.views.generic import TemplateView
from django_otp.plugins.otp_static.models import StaticToken
from django_otp.plugins.otp_totp.models import TOTPDevice

from allauth_2fa import app_settings
from allauth_2fa.forms import TOTPAuthenticateForm
from allauth_2fa.forms import TOTPDeviceForm
from allauth_2fa.forms import TOTPDeviceRemoveForm
from allauth_2fa.mixins import ValidTOTPDeviceRequiredMixin
from allauth_2fa.utils import generate_totp_config_svg
from allauth_2fa.utils import get_device_base32_secret
from allauth_2fa.utils import user_has_valid_totp_device


class TwoFactorAuthenticate(FormView):
    template_name = f"allauth_2fa/authenticate.{app_settings.TEMPLATE_EXTENSION}"
    form_class = TOTPAuthenticateForm

    def dispatch(self, request, *args, **kwargs):
        # If the user is not about to enter their two-factor credentials,
        # redirect to the login page (they shouldn't be here!). This includes
        # anonymous users.
        if "allauth_2fa_user_id" not in request.session:
            # Don't use the redirect_to_login here since we don't actually want
            # to include the next parameter.
            return redirect("account_login")
        return super().dispatch(request, *args, **kwargs)

    def get_form_class(self):
        return get_form_class(app_settings.FORMS, "authenticate", self.form_class)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        user_id = self.request.session["allauth_2fa_user_id"]
        kwargs["user"] = get_user_model().objects.get(id=user_id)
        return kwargs

    def form_valid(self, form):
        """
        The allauth 2fa login flow is now done (the user logged in successfully
        with 2FA), continue the logic from allauth.account.utils.perform_login
        since it was interrupted earlier.

        """
        adapter = get_adapter(self.request)
        # 2fa kicked in at `pre_login()`, so we need to continue from there.
        login_kwargs = adapter.unstash_pending_login_kwargs(self.request)
        adapter.login(self.request, form.user)
        return adapter.post_login(self.request, form.user, **login_kwargs)


class TwoFactorSetup(LoginRequiredMixin, FormView):
    template_name = f"allauth_2fa/setup.{app_settings.TEMPLATE_EXTENSION}"
    form_class = TOTPDeviceForm
    success_url = reverse_lazy(app_settings.SETUP_SUCCESS_URL)

    def dispatch(self, request, *args, **kwargs):
        # If the user has 2FA setup already, redirect them to the backup tokens.
        if user_has_valid_totp_device(request.user):
            return HttpResponseRedirect(self.get_success_url())

        return super().dispatch(request, *args, **kwargs)

    def _new_device(self):
        """
        Replace any unconfirmed TOTPDevices with a new one for confirmation.

        This needs to be done whenever a GET request to the page is received OR
        if the confirmation of the device fails.
        """
        self.request.user.totpdevice_set.filter(confirmed=False).delete()
        self.device = TOTPDevice.objects.create(user=self.request.user, confirmed=False)

    def get(self, request, *args, **kwargs):
        # Whenever this page is loaded, create a new device (this ensures a
        # user's QR code isn't shown multiple times).
        self._new_device()
        return super().get(request, *args, **kwargs)

    def get_qr_code_kwargs(self) -> dict[str, str]:
        """
        Get the configuration for generating a QR code.

        The fields required are:
        - `label`: identifies which account a key is associated with. Contains an
            account name, preferably prefixed by an issuer name and a colon, e.g.
            `issuer: account`.
        - `issuer`: indicates the provider or service this account is associated with.
        """

        issuer = get_current_site(self.request).name

        return {
            "issuer": issuer,
            "label": f"{issuer}: {self.request.user.get_username()}",
        }

    def get_qr_code_data_uri(self):
        svg_data = generate_totp_config_svg(
            device=self.device,
            **self.get_qr_code_kwargs(),
        )
        return f"data:image/svg+xml;base64,{force_str(b64encode(svg_data))}"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["qr_code_url"] = self.get_qr_code_data_uri()
        context["secret"] = get_device_base32_secret(self.device)
        return context

    def get_form_class(self):
        return get_form_class(app_settings.FORMS, "setup", self.form_class)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs

    def form_valid(self, form):
        # Confirm the device.
        form.save()
        return super().form_valid(form)

    def form_invalid(self, form):
        # If the confirmation code was wrong, generate a new device.
        self._new_device()
        return super().form_invalid(form)


class TwoFactorRemove(ValidTOTPDeviceRequiredMixin, FormView):
    template_name = f"allauth_2fa/remove.{app_settings.TEMPLATE_EXTENSION}"
    form_class = TOTPDeviceRemoveForm
    success_url = reverse_lazy(app_settings.REMOVE_SUCCESS_URL)

    def get_form_class(self):
        return get_form_class(app_settings.FORMS, "remove", self.form_class)

    def form_valid(self, form):
        form.save()
        return super().form_valid(form)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.request.user
        return kwargs


class TwoFactorBackupTokens(ValidTOTPDeviceRequiredMixin, TemplateView):
    template_name = f"allauth_2fa/backup_tokens.{app_settings.TEMPLATE_EXTENSION}"
    # This can be overridden in a subclass to True,
    # to have that particular view always reveal the tokens.
    reveal_tokens = bool(app_settings.ALWAYS_REVEAL_BACKUP_TOKENS)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        static_device, _ = self.request.user.staticdevice_set.get_or_create(
            name="backup",
        )

        if static_device:
            context["backup_tokens"] = static_device.token_set.all()
            context["reveal_tokens"] = self.reveal_tokens

        return context

    def post(self, request, *args, **kwargs):
        static_device, _ = request.user.staticdevice_set.get_or_create(name="backup")
        static_device.token_set.all().delete()
        for _ in range(3):
            static_device.token_set.create(token=StaticToken.random_token())
        self.reveal_tokens = True
        return self.get(request, *args, **kwargs)



from __future__ import annotations

from allauth.account import app_settings as allauth_settings
from django.conf import settings

TEMPLATE_EXTENSION = getattr(
    settings,
    "ALLAUTH_2FA_TEMPLATE_EXTENSION",
    allauth_settings.TEMPLATE_EXTENSION,
)

ALWAYS_REVEAL_BACKUP_TOKENS = bool(
    getattr(settings, "ALLAUTH_2FA_ALWAYS_REVEAL_BACKUP_TOKENS", True),
)

REMOVE_SUCCESS_URL = getattr(
    settings,
    "ALLAUTH_2FA_REMOVE_SUCCESS_URL",
    "two-factor-setup",
)

SETUP_SUCCESS_URL = getattr(
    settings,
    "ALLAUTH_2FA_SETUP_SUCCESS_URL",
    "two-factor-backup-tokens",
)

FORMS = getattr(settings, "ALLAUTH_2FA_FORMS", {})



from __future__ import annotations

from base64 import b32encode
from io import BytesIO
from urllib.parse import quote
from urllib.parse import urlencode

import qrcode
from django_otp.models import Device
from qrcode.image.svg import SvgPathImage


def get_device_base32_secret(device: Device) -> str:
    return b32encode(device.bin_key).decode("utf-8")


def generate_totp_config_svg(device: Device, issuer: str, label: str) -> bytes:
    params = {
        "secret": get_device_base32_secret(device),
        "algorithm": "SHA1",
        "digits": device.digits,
        "period": device.step,
        "issuer": issuer,
    }

    otpauth_url = f"otpauth://totp/{quote(label)}?{urlencode(params)}"

    img = qrcode.make(otpauth_url, image_factory=SvgPathImage)
    io = BytesIO()
    img.save(io)
    return io.getvalue()


def user_has_valid_totp_device(user) -> bool:
    if not user.is_authenticated:
        return False
    return user.totpdevice_set.filter(confirmed=True).exists()



from __future__ import annotations

from django.urls import path

from allauth_2fa import views

urlpatterns = [
    path(
        "authenticate/",
        views.TwoFactorAuthenticate.as_view(),
        name="two-factor-authenticate",
    ),
    path(
        "setup/",
        views.TwoFactorSetup.as_view(),
        name="two-factor-setup",
    ),
    path(
        "backup-tokens/",
        views.TwoFactorBackupTokens.as_view(),
        name="two-factor-backup-tokens",
    ),
    path(
        "remove/",
        views.TwoFactorRemove.as_view(),
        name="two-factor-remove",
    ),
]



from __future__ import annotations

from allauth.account.adapter import get_adapter
from django.conf import settings
from django.contrib import messages
from django.http import HttpRequest
from django.http import HttpResponse
from django.shortcuts import redirect
from django.urls import resolve
from django.utils.deprecation import MiddlewareMixin


class AllauthTwoFactorMiddleware(MiddlewareMixin):
    """
    Reset the login flow if another page is loaded halfway through the login.
    (I.e. if the user has logged in with a username/password, but not yet
    entered their two-factor credentials.) This makes sure a user does not stay
    half logged in by mistake.

    """

    def process_request(self, request: HttpRequest) -> None:
        match = resolve(request.path)
        if not match.url_name or not match.url_name.startswith(
            "two-factor-authenticate",
        ):
            try:
                del request.session["allauth_2fa_user_id"]
            except KeyError:
                pass


class BaseRequire2FAMiddleware(MiddlewareMixin):
    """
    Ensure that particular users have two-factor authentication enabled before
    they have access to the rest of the app.

    If they don't have 2FA enabled, they will be redirected to the 2FA
    enrollment page and not be allowed to access other pages.
    """

    # List of URL names that the user should still be allowed to access.
    allowed_pages = [
        # They should still be able to log out or change password.
        "account_change_password",
        "account_logout",
        "account_reset_password",
        # URLs required to set up two-factor
        "two-factor-setup",
    ]
    # The message to the user if they don't have 2FA enabled and must enable it.
    require_2fa_message = (
        "You must enable two-factor authentication before doing anything else."
    )

    def on_require_2fa(self, request: HttpRequest) -> HttpResponse:
        """
        If the current request requires 2FA and the user does not have it
        enabled, this is executed. The result of this is returned from the
        middleware.
        """
        # See allauth.account.adapter.DefaultAccountAdapter.add_message.
        if "django.contrib.messages" in settings.INSTALLED_APPS:
            # If there is already a pending message related to two-factor (likely
            # created by a redirect view), simply update the message text.
            storage = messages.get_messages(request)
            tag = "2fa_required"
            for m in storage:
                if m.extra_tags == tag:
                    m.message = self.require_2fa_message
                    break
            # Otherwise, create a new message.
            else:
                messages.error(request, self.require_2fa_message, extra_tags=tag)
            # Mark the storage as not processed so they'll be shown to the user.
            storage.used = False

        # Redirect user to two-factor setup page.
        return redirect("two-factor-setup")

    def require_2fa(self, request: HttpRequest) -> bool:
        """
        Check if this request is required to have 2FA before accessing the app.

        This should return True if this request requires 2FA.

        You can access anything on the request, but generally request.user will
        be most interesting here.
        """
        raise NotImplementedError("You must implement require_2fa.")

    def is_allowed_page(self, request: HttpRequest) -> bool:
        return request.resolver_match.url_name in self.allowed_pages

    def process_view(
        self,
        request: HttpRequest,
        view_func,
        view_args,
        view_kwargs,
    ) -> HttpResponse | None:
        # The user is not logged in, do nothing.
        if request.user.is_anonymous:
            return None

        # If this doesn't require 2FA, then stop processing.
        if not self.require_2fa(request):
            return None

        # If the user is on one of the allowed pages, do nothing.
        if self.is_allowed_page(request):
            return None

        # User already has two-factor configured, do nothing.
        if get_adapter(request).has_2fa_enabled(request.user):
            return None

        # The request required 2FA but it isn't configured!
        return self.on_require_2fa(request)



from __future__ import annotations

import contextlib

from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext_lazy as _
from django_otp.forms import OTPAuthenticationFormMixin
from django_otp.plugins.otp_totp.models import TOTPDevice

DEFAULT_TOKEN_WIDGET_ATTRS = {
    "autofocus": "autofocus",
    "autocomplete": "off",
    "inputmode": "numeric",
}


class _TokenToOTPTokenMixin:
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if "token" in self.fields or "token" in self.data:
            self._raise_token_exception()

    @property
    def token(self):
        self._raise_token_exception()

    def _raise_token_exception(self):
        raise ImproperlyConfigured(
            f"The field `token` in {self} has been renamed to `otp_token`.",
        )


class TOTPAuthenticateForm(OTPAuthenticationFormMixin, forms.Form):
    otp_token = forms.CharField(
        label=_("Token"),
    )

    def __init__(self, user, **kwargs):
        super().__init__(**kwargs)
        self.fields["otp_token"].widget.attrs.update(DEFAULT_TOKEN_WIDGET_ATTRS)
        self.user = user

    def clean(self) -> dict:
        self.clean_otp(self.user)
        return self.cleaned_data


class TOTPDeviceForm(_TokenToOTPTokenMixin, forms.Form):
    otp_token = forms.CharField(
        label=_("Token"),
    )

    def __init__(self, user, metadata=None, **kwargs):
        super().__init__(**kwargs)
        self.fields["otp_token"].widget.attrs.update(DEFAULT_TOKEN_WIDGET_ATTRS)
        self.user = user
        self.metadata = metadata or {}

    def clean_otp_token(self):
        token = self.cleaned_data.get("otp_token")

        # Find the unconfirmed device and attempt to verify the token.
        self.device = self.user.totpdevice_set.filter(confirmed=False).first()
        if not self.device.verify_token(token):
            raise forms.ValidationError(_("The entered token is not valid"))

        return token

    def save(self) -> TOTPDevice:
        # The device was found to be valid, delete other confirmed devices and
        # confirm the new device.
        self.user.totpdevice_set.filter(confirmed=True).delete()
        self.device.confirmed = True
        self.device.save()

        return self.device


class TOTPDeviceRemoveForm(
    _TokenToOTPTokenMixin,
    OTPAuthenticationFormMixin,
    forms.Form,
):
    # User must input a valid token so 2FA can be removed
    otp_token = forms.CharField(
        label=_("Token"),
    )

    def __init__(self, user, **kwargs):
        super().__init__(**kwargs)

        self.user = user
        self.fields["otp_token"].widget.attrs.update(DEFAULT_TOKEN_WIDGET_ATTRS)

    def clean(self):
        self.clean_otp(self.user)
        return self.cleaned_data

    def save(self) -> None:
        with contextlib.suppress(ObjectDoesNotExist):
            # Delete any backup tokens and their related static device.
            static_device = self.user.staticdevice_set.get(name="backup")
            static_device.token_set.all().delete()
            static_device.delete()

        # Delete TOTP device.
        device = TOTPDevice.objects.get(user=self.user)
        device.delete()
